#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdarg.h>
#include <time.h>
#include <string.h>
#include <assert.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/ioctl.h>

#include "uart16550.h"

#define UART16550_MAJOR 42
#define COM2_MAJOR 43
#define COM1_MAJOR 42
#define STR(x)        #x
#define XSTR(x)       STR(x)
#define OPTION_BOTH		3
#define OPTION_COM1_ONLY	1
#define OPTION_COM2_ONLY	2


static int testno=0;

#define BUFFER_SIZE 256

#define test(t, s...) \
({\
	int i, output_chars=0,v;\
	\
        output_chars=printf("test %04d: ", ++testno); \
	output_chars+=printf(s); \
	fflush(stdout);\
	for(i=0; i<60-output_chars; i++)\
		putchar('.');\
	if (!(v=(t)))\
	        printf("failed\n");\
	else\
		printf("passed\n");\
	fflush(stdout);\
	v;\
})

#define fatal_test(t, s...) if (!test(t, s)) abort()

int do_read(int fd, unsigned char *buffer, int size)
{
	int n, from=0;
	
	while (1) {
		n=read(fd, &buffer[from], size-from);
		if (n <= 0)
		        return -1;
	        if (n+from == size)
	    	        return 0;
		from+=n;
	}
}

int do_write(int fd, unsigned char *buffer, unsigned int size)
{
	int n, from=0;
	
	while (1) {
		n=write(fd, &buffer[from], size-from);
		if (n <= 0)
		        return -1;
	        if (n+from == size)
	    	        return 0;
		from+=n;
	}
}


void test_group(const char *com1_path, const char *com2_path, int baud, int loops)
{
	int com1, com2, i, j;
	struct uart16550_line_info uli = { baud, UART16550_LEN_8, UART16550_PAR_STICK, UART16550_STOP_2};
	unsigned char wb[BUFFER_SIZE], rb[BUFFER_SIZE];	
	time_t start, stop;


	start=time(NULL);
	test(open("/dev/uart10", O_RDWR) == -1 && errno == ENXIO, "open /dev/uart10");
	fatal_test((com1=open(com1_path, O_RDWR)) >= 0, "open %s", com1_path);
	fatal_test((com2=open(com2_path, O_RDWR)) >= 0, "open %s", com2_path);
	test(ioctl(com1, 0xdeadbeef, &uli) != 0, "dummy ioctl %s", com1_path);  
	test(ioctl(com2, 0xdeadbeef, &uli) != 0, "dummy ioctl %s", com2_path);  
	test(ioctl(com1, UART16550_IOCTL_SET_LINE, &uli) == 0, "set line %s", com1_path);  
	test(ioctl(com2, UART16550_IOCTL_SET_LINE, &uli) == 0, "set line %s", com2_path);  
	for(j=0; j<loops; j++) {
		for(i=0; i<BUFFER_SIZE; i++) {
			wb[i]=rand()%256;
			rb[i]=0;
		}
		test(do_write(com1, wb, BUFFER_SIZE) == 0, "write %s", com1_path);
		test(do_read(com2, rb, BUFFER_SIZE) == 0, "read %s", com2_path);
		for(i=0; i<BUFFER_SIZE; i++) {
			if (rb[i] != wb[i])
				break;
	    	}
	    	if (i >= 16)
			test(1, "data check");
		else {
			for(i=0; i<BUFFER_SIZE; i++)
				if (rb[i] != wb[i])
					fprintf(stderr, "%d %x %x\n", i, rb[i], wb[i]);
			test(0, "data check");
		}
	}
	test(close(com1) == 0, "close %s", com1_path);
	test(close(com2) == 0, "close %s", com2_path);
	stop=time(NULL);

	test(stop-start <= 15, "time (%lds) <= 15s", stop-start); 
}

int main(int argc, char **argv)
{
	srand(time(NULL));

	mknod("/dev/uart0", S_IFCHR, COM1_MAJOR<<8);
	mknod("/dev/uart1", S_IFCHR, COM2_MAJOR<<8);
	mknod("/dev/uart10", S_IFCHR, (UART16550_MAJOR<<8)+10);

	fprintf(stderr,"==== Module insertion and port&interrupt registration ============\n");

	fatal_test(system("insmod uart16550.ko major=" XSTR(COM1_MAJOR) " option=" XSTR(OPTION_COM1_ONLY)) == 0, 
			"insmod uart16550.ko major=" XSTR(COM1_MAJOR));
	fatal_test(system("insmod solutie.ko major=" XSTR(COM2_MAJOR) " option=" XSTR(OPTION_COM2_ONLY)) == 0, 
			"insmod solutie.ko major=" XSTR(COM2_MAJOR));

	test(system("cat /proc/ioports | grep \'03f8-03ff : uart16550\' > /dev/null 2>&1") == 0, "COM1 I/O ports");
	test(system("cat /proc/ioports | grep \'02f8-02ff : uart16550\' > /dev/null 2>&1") == 0, "COM2 I/O ports");
	fatal_test(system("cat /proc/interrupts | grep \'4:.*uart16550\' > /dev/null 2>&1") == 0, "COM1 irq");
	fatal_test(system("cat /proc/interrupts | grep \'3:.*uart16550\' > /dev/null 2>&1") == 0, "COM2 irq");
	fatal_test( system("rmmod uart16550") == 0, "rmmod uart16550");
	fatal_test( system("rmmod solutie") == 0, "rmmod solutie");
	test(system("cat /proc/ioports | grep \"03f8-03ff : uart16550\" > /dev/null 2>&1") != 0, "COM1 I/O ports");
	test(system("cat /proc/ioports | grep \"02f8-02ff : uart16550\" > /dev/null 2>&1") != 0, "COM2 I/O ports");
	fatal_test(system("cat /proc/interrupts | grep \"4:.*uart16550\" > /dev/null 2>&1") != 0, "COM1 irq");
	fatal_test(system("cat /proc/interrupts | grep \"3:.*uart16550\" > /dev/null 2>&1") != 0, "COM2 irq");


	fprintf(stderr,"==== Reading at base speed (2400 baud) ===========================\n");

	fatal_test(system("insmod uart16550.ko major=" XSTR(COM1_MAJOR) " option=" XSTR(OPTION_COM1_ONLY)) == 0, 
			"insmod uart16550.ko major=" XSTR(COM1_MAJOR));
	fatal_test(system("insmod solutie.ko major=" XSTR(COM2_MAJOR) " option=" XSTR(OPTION_COM2_ONLY)) == 0, 
			"insmod solutie.ko major=" XSTR(COM2_MAJOR));

	test(system("cat /proc/ioports | grep \'03f8-03ff : uart16550\' > /dev/null 2>&1") == 0, "COM1 I/O ports");
	test(system("cat /proc/ioports | grep \'02f8-02ff : uart16550\' > /dev/null 2>&1") == 0, "COM2 I/O ports");
	fatal_test(system("cat /proc/interrupts | grep \'4:.*uart16550\' > /dev/null 2>&1") == 0, "COM1 irq");
	fatal_test(system("cat /proc/interrupts | grep \'3:.*uart16550\' > /dev/null 2>&1") == 0, "COM2 irq");
	test_group("/dev/uart1", "/dev/uart0", UART16550_BAUD_2400, 5);
	fatal_test( system("rmmod uart16550") == 0, "rmmod uart16550");
	fatal_test( system("rmmod solutie") == 0, "rmmod solutie");
	test(system("cat /proc/ioports | grep \"03f8-03ff : uart16550\" > /dev/null 2>&1") != 0, "COM1 I/O ports");
	test(system("cat /proc/ioports | grep \"02f8-02ff : uart16550\" > /dev/null 2>&1") != 0, "COM2 I/O ports");
	fatal_test(system("cat /proc/interrupts | grep \"4:.*uart16550\" > /dev/null 2>&1") != 0, "COM1 irq");
	fatal_test(system("cat /proc/interrupts | grep \"3:.*uart16550\" > /dev/null 2>&1") != 0, "COM2 irq");

	fprintf(stderr,"==== Writing at base speed (2400 baud) ===========================\n");

	fatal_test(system("insmod uart16550.ko major=" XSTR(COM2_MAJOR) " option=" XSTR(OPTION_COM2_ONLY)) == 0, 
			"insmod uart16550.ko major=" XSTR(COM2_MAJOR));
	fatal_test(system("insmod solutie.ko major=" XSTR(COM1_MAJOR) " option=" XSTR(OPTION_COM1_ONLY)) == 0, 
			"insmod solutie.ko major=" XSTR(COM1_MAJOR));

	test(system("cat /proc/ioports | grep \'03f8-03ff : uart16550\' > /dev/null 2>&1") == 0, "COM1 I/O ports");
	test(system("cat /proc/ioports | grep \'02f8-02ff : uart16550\' > /dev/null 2>&1") == 0, "COM2 I/O ports");
	fatal_test(system("cat /proc/interrupts | grep \'4:.*uart16550\' > /dev/null 2>&1") == 0, "COM1 irq");
	fatal_test(system("cat /proc/interrupts | grep \'3:.*uart16550\' > /dev/null 2>&1") == 0, "COM2 irq");
	test_group("/dev/uart1", "/dev/uart0", UART16550_BAUD_2400, 5);
	fatal_test( system("rmmod uart16550") == 0, "rmmod uart16550");
	fatal_test( system("rmmod solutie") == 0, "rmmod solutie");
	test(system("cat /proc/ioports | grep \"03f8-03ff : uart16550\" > /dev/null 2>&1") != 0, "COM1 I/O ports");
	test(system("cat /proc/ioports | grep \"02f8-02ff : uart16550\" > /dev/null 2>&1") != 0, "COM2 I/O ports");
	fatal_test(system("cat /proc/interrupts | grep \"4:.*uart16550\" > /dev/null 2>&1") != 0, "COM1 irq");
	fatal_test(system("cat /proc/interrupts | grep \"3:.*uart16550\" > /dev/null 2>&1") != 0, "COM2 irq");

	/*
	test_group("/dev/uart0", "/dev/uart1", UART16550_BAUD_2400, 1);
	test_group("/dev/uart1", "/dev/uart0", UART16550_BAUD_2400, 1);
	*/
	fprintf(stderr,"==== Reading/Writing tests at different speeds ==================\n");

	fatal_test(system("insmod uart16550.ko major=" XSTR(COM1_MAJOR) " option=" XSTR(OPTION_COM1_ONLY)) == 0, 
			"insmod uart16550.ko major=" XSTR(COM1_MAJOR));
	fatal_test(system("insmod solutie.ko major=" XSTR(COM2_MAJOR) " option=" XSTR(OPTION_COM2_ONLY)) == 0, 
			"insmod solutie.ko major=" XSTR(COM2_MAJOR));

	test(system("cat /proc/ioports | grep \'03f8-03ff : uart16550\' > /dev/null 2>&1") == 0, "COM1 I/O ports");
	test(system("cat /proc/ioports | grep \'02f8-02ff : uart16550\' > /dev/null 2>&1") == 0, "COM2 I/O ports");
	fatal_test(system("cat /proc/interrupts | grep \'4:.*uart16550\' > /dev/null 2>&1") == 0, "COM1 irq");
	fatal_test(system("cat /proc/interrupts | grep \'3:.*uart16550\' > /dev/null 2>&1") == 0, "COM2 irq");
	test_group("/dev/uart0", "/dev/uart1", UART16550_BAUD_9600, 20);
	test_group("/dev/uart1", "/dev/uart0", UART16550_BAUD_9600, 20);
	fatal_test( system("rmmod uart16550") == 0, "rmmod uart16550");
	fatal_test( system("rmmod solutie") == 0, "rmmod solutie");
	test(system("cat /proc/ioports | grep \"03f8-03ff : uart16550\" > /dev/null 2>&1") != 0, "COM1 I/O ports");
	test(system("cat /proc/ioports | grep \"02f8-02ff : uart16550\" > /dev/null 2>&1") != 0, "COM2 I/O ports");
	fatal_test(system("cat /proc/interrupts | grep \"4:.*uart16550\" > /dev/null 2>&1") != 0, "COM1 irq");
	fatal_test(system("cat /proc/interrupts | grep \"3:.*uart16550\" > /dev/null 2>&1") != 0, "COM2 irq");

	fatal_test(system("insmod uart16550.ko major=" XSTR(COM1_MAJOR) " option=" XSTR(OPTION_COM1_ONLY)) == 0, 
			"insmod uart16550.ko major=" XSTR(COM1_MAJOR));
	fatal_test(system("insmod solutie.ko major=" XSTR(COM2_MAJOR) " option=" XSTR(OPTION_COM2_ONLY)) == 0, 
			"insmod solutie.ko major=" XSTR(COM2_MAJOR));

	test(system("cat /proc/ioports | grep \'03f8-03ff : uart16550\' > /dev/null 2>&1") == 0, "COM1 I/O ports");
	test(system("cat /proc/ioports | grep \'02f8-02ff : uart16550\' > /dev/null 2>&1") == 0, "COM2 I/O ports");
	fatal_test(system("cat /proc/interrupts | grep \'4:.*uart16550\' > /dev/null 2>&1") == 0, "COM1 irq");
	fatal_test(system("cat /proc/interrupts | grep \'3:.*uart16550\' > /dev/null 2>&1") == 0, "COM2 irq");
	test_group("/dev/uart0", "/dev/uart1", UART16550_BAUD_115200, 20);
	test_group("/dev/uart1", "/dev/uart0", UART16550_BAUD_115200, 20);
	test_group("/dev/uart0", "/dev/uart1", UART16550_BAUD_115200, 20);
	test_group("/dev/uart1", "/dev/uart0", UART16550_BAUD_115200, 20);
	fatal_test( system("rmmod uart16550") == 0, "rmmod uart16550");
	fatal_test( system("rmmod solutie") == 0, "rmmod solutie");
	test(system("cat /proc/ioports | grep \"03f8-03ff : uart16550\" > /dev/null 2>&1") != 0, "COM1 I/O ports");
	test(system("cat /proc/ioports | grep \"02f8-02ff : uart16550\" > /dev/null 2>&1") != 0, "COM2 I/O ports");
	fatal_test(system("cat /proc/interrupts | grep \"4:.*uart16550\" > /dev/null 2>&1") != 0, "COM1 irq");
	fatal_test(system("cat /proc/interrupts | grep \"3:.*uart16550\" > /dev/null 2>&1") != 0, "COM2 irq");

	unlink("/dev/uart0");
	unlink("/dev/uart1");
	unlink("/dev/uart10");

	return 0;
}

